#!/usr/bin/env python
"""
Scummbler v2

    This file is part of Scummbler.

    Scummbler is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Scummbler is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Scummbler.  If not, see <http://www.gnu.org/licenses/>.

Laurence Dougal Myers
www.jestarjokin.net

December 2008 - December 2010

Atomic parsing elements
"""
from pyparsing import *
import scummbler_vars

# le = lexicon entry, also gives the code a French feel

# Numbers
leUnsignedInt = Word(nums)
leInt = Word(nums + "-", nums)
leHexByte = Word(nums + srange("[a-fA-F]"), exact=2) # for descummed scripts
leHexWord = Word(nums + srange("[a-fA-F]"), exact=4) # for descummed scripts
leHexInt = Suppress(Literal("0x")) + Word(nums + srange("[a-fA-F]"), max=8) # 32-bit
leDefinedValueDef = Word(alphanums + "_") # prevents parsing definedValues on definition
#leDefinedValue = Word("$", alphanums + "_") # compiler define value (I hate using the $)
leDefinedValue = Forward() # the compiler has to fill this in
leByte = (leInt | leDefinedValue)("value")
leWord = (leInt | leDefinedValue)("value")
leDelay = leUnsignedInt
leLabel = Word(alphanums + "_")
leLocalIndex = leUnsignedInt
leBitIndex = leUnsignedInt # same for now
leGlobalIndex = leUnsignedInt # same for now
leLocalIndexOrPointer = Forward() # I'm not sure if this can be indirect
leBitIndexOrPointer = Forward() # I'm not sure if this can be indirect
leGlobalIndexOrPointer = Forward()

# Variables (these could probably be condensed)
leLocalVar = Literal("Local")("vartype") + Suppress(Literal("[")) + leLocalIndexOrPointer("value") + Suppress(Literal("]"))
leBitVar = Literal("Bit")("vartype") + Suppress(Literal("[")) + leBitIndexOrPointer("value") + Suppress(Literal("]"))
leGlobalVar = Literal("Var")("vartype") + Suppress(Literal("[")) + leGlobalIndexOrPointer("value") + Suppress(Literal("]"))
leKnownGlobalVar = Forward()

leVariable = leLocalVar | leBitVar | leGlobalVar | leKnownGlobalVar | leDefinedValue
leVarOrByte = leVariable | leByte
leVarOrWord = leVariable | leWord
leListItem = Group(leVariable | leWord) # group is a hack to work around nested vars etc

leIndexPointer = Group(leUnsignedInt("base") + Literal("+") + Group(leVarOrWord)("offset"))

# Numbers (Forwarded/Recursive)
leLocalIndexOrPointer << (leIndexPointer | leLocalIndex) # removed because only globals can be indirect
leBitIndexOrPointer << (leIndexPointer | leBitIndex) # removed because only globals can be indirect
leGlobalIndexOrPointer << (leIndexPointer | leGlobalIndex)

leArgList = Suppress(Literal("[")) + Optional(delimitedList(leListItem))("list") + Suppress(Literal("]"))

# Other
# These string functions probably shouldn't be in the lexicon... baaaahh!
leStringFuncNewline = Literal("newline")("function") + Suppress(Literal("(")) + Suppress(Literal(")"))
leStringFuncKeepText = Literal("keepText")("function") + Suppress(Literal("(")) + Suppress(Literal(")"))
leStringFuncWait = Literal("wait")("function") + Suppress(Literal("(")) + Suppress(Literal(")"))
leStringFuncGetInt = Literal("getInt")("function") + Suppress(Literal("(")) + Group(leVariable)("arg1") + Suppress(Literal(")"))
leStringFuncGetVerb = Literal("getVerb")("function") + Suppress(Literal("(")) + Group(leVariable)("arg1") + Suppress(Literal(")"))
leStringFuncGetName = Literal("getName")("function") + Suppress(Literal("(")) + Group(leVariable)("arg1") + Suppress(Literal(")"))
leStringFuncGetString = Literal("getString")("function") + Suppress(Literal("(")) + Group(leVariable)("arg1") + Suppress(Literal(")"))
leStringFuncStartAnim = Literal("startAnim")("function") + Suppress(Literal("(")) + leWord("arg1") + Suppress(Literal(")"))
leStringFuncSound = (
    Literal("sound")("function") + Suppress(Literal("(")) +
        (leHexInt | leInt)("arg1") + Suppress(Literal(",")) +
        (leHexInt | leInt)("arg2") +
    Suppress(Literal(")"))
)
leStringFuncSetColor = Literal("setColor")("function") + Suppress(Literal("(")) + leWord("arg1") + Suppress(Literal(")"))
leStringFuncSetFont = Literal("setFont")("function") + Suppress(Literal("(")) + leWord("arg1") + Suppress(Literal(")"))
# TODO: support "unknown" functions
leStringFunc = (leStringFuncNewline | leStringFuncKeepText | leStringFuncWait |
                leStringFuncGetInt | leStringFuncGetVerb | leStringFuncGetName | leStringFuncGetString |
                leStringFuncStartAnim | leStringFuncSound | leStringFuncSetColor | leStringFuncSetFont)
leStringText = quotedString

leStringParticle = leStringText | leStringFunc
leString = leStringParticle + Optional(OneOrMore(Suppress(Literal("+")) + leStringParticle))


leComment = cStyleComment # luckily, descumm uses C-style comments!
